/*****************************************************************************
 * LEAP implementation.
 *
 * Licensed under a dual GPL/BSD license.  (See LICENSE file for more info.)
 *
 * File: eapleap.c
 *
 * Authors: Marios Karagiannopoulos (marios@master.math.upatras.gr)
 *
 * Modified to support dynamic keying by Chris Hessing, with help from
 * Gilbert Goodwill.
 *
 ****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "xsup_debug.h"
#include "xsup_err.h"
#include "frame_structs.h"

#include "xsupconfig.h"
#include "profile.h"
#include "eap.h"
#include "eapleap.h"
#include "leapmd4.h"
#include "eap_types/mschapv2/mschapv2.h"

#ifdef USE_EFENCE
#include <efence.h>
#endif

#define LEAP_LENGTH    0x08
struct leap_requests *leaprequest;
struct leap_responses *leapresponse;
struct leap_challenges *leapchallenges;

static void ntPwdHash(unsigned char *MD4Hash, char *password) {
    char unicodePass[513];
    char passLen;
    int i;

    if (!xsup_assert((MD4Hash != NULL), "MD4Hash != NULL", FALSE))
      return;

    if (!xsup_assert((password != NULL), "password != NULL", FALSE))
      return;

    /* Microsoft passwords are unicode.  Convert plain text password
       to unicode by inserting a zero every other byte */
    passLen = strlen(password);
    for (i = 0; i < passLen; i++) {
        unicodePass[2 * i] = password[i];
        unicodePass[2 * i + 1] = 0;
    }
    /* Encrypt plain text password to a 16-byte MD4 hash */
    md4_calc(MD4Hash, (uint8_t *) unicodePass, passLen * 2);
}

void leap_mschap(char * password, char * response) {
    unsigned char MD4Hash[16], MD4HashHash[16];

    if (!xsup_assert((password != NULL), "password != NULL", FALSE))
      return;

    if (!xsup_assert((response != NULL), "response != NULL", FALSE))
      return;

    ntPwdHash(MD4Hash, password);
    md4_calc(MD4HashHash, MD4Hash, 16);
    ChallengeResponse((char *) leapchallenges->apc, (char *) MD4HashHash, response);
}


/*****************************************************
 *
 * Setup to handle LEAP EAP requests
 *
 * This function is called each time we receive a packet of the EAP type LEAP.
 * At a minimum, it should check to make sure it's stub in the structure
 * exists, and if not, set up any variables it may need.  Since LEAP doesn't
 * have any state that needs to survive successive calls, we don't need to
 * do anything here.
 *
 *****************************************************/
int eapleap_setup(struct generic_eap_data *thisint)
{
  struct leap_data *mydata;

  if (!xsup_assert((thisint != NULL), "thisint != NULL", FALSE))
    return XEMALLOC;

  mydata = (struct leap_data *)malloc(sizeof(struct leap_data));
  if (mydata == NULL) 
    {
      debug_printf(DEBUG_NORMAL, "Cannot allocate memory in eapleap_setup()!\n");
      return XEMALLOC;
    }

  mydata->keyingMaterial = NULL;
  mydata->eapsuccess = FALSE;

  thisint->eap_data = mydata;
 
  return XENONE;
}

/*************************************************************
  leap_decode_packet - decode an LEAP challenge, and answer it

  	Cisco LEAP authenticates users to the wireless access point via a
password.  This password is authenticated against a back-end radius server
via a Challenge-Response protocol.  The protocol is such:
	1.) The Wireless client sends an authentication request;
	2.) The AP Acknowledges request with an 8 byte challenge;
	3.) The Wireless client computes the response by:
		a.) MD4 Hashing the password producing a 16 byte hash;
		b.) Padding the hash with 5 nulls producing 21 bytes;
		c.) Splitting the resulting 21 bytes into 7 byte chunks;
		d.) Iterating through the 7 byte chunks, des encrypting
			the challenge as plain-text with the 7-byte chunk
			as the key.
		e.) Concatenating the resulting cipher text producing 24
			bytes
	4.) The client then sends the resulting 24 bytes as the challenge
		response;
	5.) The back-end systems iterate through the same processes and
		check for a match; then
	6.) If the two match, authentication has been accomplished.

************************************************************/

int eapleap_process(struct generic_eap_data *thisint, uint8_t *dataoffs,
		   int insize, uint8_t *outframe, int *outsize)

{
  struct eap_header *eapheader;
  char *answer = NULL;
  char *data, *username;
  unsigned char chall_response[24];
  int total_length;
  unsigned char MD4Hash[16], MD4HashHash[16];
  char MasterKey[16];
  struct config_eap_leap *userdata;
  struct leap_data *mydata;
  unsigned char challenge_response_expected[24];
  unsigned char *challenge_response_got;

  if (!xsup_assert((thisint != NULL), "thisint != NULL", FALSE))
    return XEMALLOC;

  if (!xsup_assert((dataoffs != NULL), "dataoffs != NULL", FALSE))
    return XEMALLOC;

  if (!xsup_assert((outframe != NULL), "outframe != NULL", FALSE))
    return XEMALLOC;

  if (!xsup_assert((outsize != NULL), "outsize != NULL", FALSE))
    return XEMALLOC;

  debug_printf(DEBUG_EVERYTHING, "Processing LEAP\n");

  if (!xsup_assert((thisint->eap_conf_data != NULL), 
		   "thisint->eap_conf_data != NULL", FALSE))
    return XEMALLOC;

  userdata = (struct config_eap_leap *)thisint->eap_conf_data;

  if (!xsup_assert((userdata != NULL), "userdata != NULL", FALSE))
    return XEMALLOC;

  if (!xsup_assert((thisint->eap_data != NULL), "thisint->eap_data != NULL",
		   FALSE))
    return XEMALLOC;

  mydata = (struct leap_data *)thisint->eap_data;

  if ((thisint->tempPwd == NULL) && (userdata->password == NULL))
    {
      thisint->need_password = 1;
      thisint->eaptype = strdup("LEAP");
      thisint->eapchallenge = NULL;
      *outsize = 0;

      return XENONE;
    }

  // Make sure we have something to process...
  if (dataoffs == NULL) return XENONE;

  if (userdata->username == NULL)
    {
      username = thisint->identity;
    } else {
      username = userdata->username;
    }

  // Actually process, and respond to challenges.

  // LEAP shouldn't be used as an inner type, so we should be able to get
  // away with this.
  data = (char *) dataoffs-5;
  eapheader = (struct eap_header *)data;

  switch (eapheader->eap_code) {

  case EAP_REQUEST:
      // ***********************************************************************************************
    mydata->eapsuccess = FALSE;

    leapchallenges = (struct leap_challenges *)malloc(sizeof(struct leap_challenges));
    leaprequest = (struct leap_requests *)malloc(sizeof(struct leap_requests));

    // extract the payload received
    memcpy((struct leap_requests *)leaprequest, (struct leap_requests *)dataoffs, 16);

    // store Peer Challenge
    memcpy((uint8_t *)leapchallenges->pc, (char *)leaprequest->randval, 8);

    if (leaprequest->count != LEAP_LENGTH)  {
      debug_printf(DEBUG_NORMAL, "(EAP-LEAP) Incorrect length value for LEAP random value.\n");
      return XELEAP;
    }
    memset(chall_response, 0x0, 24);

    // Get our username and password out of our configuration structure in memory
    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) ID : %d\n",eapheader->eap_identifier);
    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) Username = %s   --   Password = %s\n", username,userdata->password);
    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) Incoming Peer Challenge Random Value (Length = %d) : ",leaprequest->count);
    debug_hex_printf(DEBUG_AUTHTYPES,(uint8_t *)leaprequest->randval, leaprequest->count);

    NtChallengeResponse((char *)leapchallenges->pc, userdata->password, 
			(char *)&chall_response, 0);

    debug_printf(DEBUG_AUTHTYPES, "MSCHAP Response Calculated : ");
    debug_hex_printf(DEBUG_AUTHTYPES, (uint8_t *)&chall_response, 24);

     // store Peer Response
    memcpy((uint8_t *)leapchallenges->pr, (char *)chall_response, 24);

    total_length = 24+3+strlen(username)+1;
    answer = (char *)malloc(total_length);
    if (answer == NULL) {
        debug_printf(DEBUG_NORMAL, "(EAP-LEAP) Couldn't allocate memory for building hash source!\n");
        return XEMALLOC;
    }
    // Construct the LEAP response sub fields packet
    // let's start with the version number (LEAP subfield)

    // byte 0: Version
    // byte 1: Unused - Reserved
    // byte 2: Count
    // byte 3..26: MS-CHAP Challenge Response
    // byte 27..m: username

    answer[0] = 0x01;
    answer[1] = 0x00; // Reserved - Unused
    answer[2] = 24; // Count

    // Include MSCHAP Challenge response in the built packet
    memcpy(&answer[3],&chall_response,24);
    // Include username in the built packet
    memcpy(&answer[24+3],username,strlen(username)+1);
    // be sure that the username (last field) will be NUL terminated!
    answer[strlen(answer)] = '\0';

    // Set up our response frame.
    memcpy(outframe, answer, total_length);
    *outsize = total_length;

    if (answer != NULL)
      free(answer);
    answer=NULL;

  break;

  case EAP_SUCCESS:
    // ***********************************************************************************************
    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) Got EAP-SUCCESS\n");
    memset(chall_response, 0x0, 8);
    NtChallengeResponse((char *)leaprequest->randval, userdata->password, 
			(char *)&chall_response, 0);

    GenerateNTResponse((char *)leapchallenges->pr, (char *)leapchallenges->pc, 
		       username, userdata->password, (char *) chall_response, 
		       0);

    // store Access Point Challenge
    memcpy((uint8_t *)leapchallenges->apc, (char *)chall_response, 8);

    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) GenerateNTResponse Calculated : ");
    debug_hex_printf(DEBUG_AUTHTYPES, (uint8_t *)&chall_response, 8);

    total_length = 8+3+strlen(username)+1;
    answer = (char *)malloc(total_length);
    if (answer == NULL) {
        debug_printf(DEBUG_NORMAL, "(EAP-LEAP) Couldn't allocate memory for building hash source!\n");
        return XEMALLOC;
    }

    // Construct the LEAP request sub fields packet
    // let's start with the version number (LEAP subfield)

    // byte 0: Version
    // byte 1: Unused - Reserved
    // byte 2: Count
    // byte 3..10: MS-CHAP Nt Challenge Response
    // byte 11..m: username

    answer[0] = 0x01;
    answer[1] = 0x00; // Reserved - Unused
    answer[2] = 8; // Count

    // Include MSCHAP Challenge response in the built packet
    memcpy(&answer[3],&chall_response,8);
    // Include username in the built packet
    memcpy(&answer[8+3],username,strlen(username)+1);
    // be sure that the username (last field) will be NUL terminated!
    answer[strlen(answer)] = '\0';

    // Set up our response frame.
    memcpy(outframe, answer, total_length);
    *outsize = total_length;

    // Store the new random value to the leapdata for further validation of the AP response !
    memcpy((char *)leaprequest->randval, (char *)&chall_response,8);

    if (answer != NULL)
      free(answer);
    answer=NULL;

    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) Request Packet for Mutual Authentication Built\n");

  break;

  case EAP_RESPONSE:
    // ***********************************************************************************************
    // Verify an AP-Challenge Response from an EAP LEAP response frame.
    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) Got EAP-RESPONSE\n");
    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) Verification phase....\n");

    leapresponse = (struct leap_responses *)dataoffs;

    challenge_response_got =  (uint8_t *)malloc(leapresponse->count+1);
    if (!challenge_response_got) {
      debug_printf(DEBUG_NORMAL, "(EAP-LEAP) challenge_response_got is NULL after malloc!\n");
    }
    memcpy(challenge_response_got, &leapresponse->randval, leapresponse->count);

    // store Access Point Response
    memcpy((uint8_t *)leapchallenges->apr, (char *)leapresponse->randval, 24);

    // Let's construct the expected one
    memset(challenge_response_expected, 0x0, 24);

    // Calculate the 24 bytes MS-CHAP Challenge Response
    leap_mschap(userdata->password, (char *) challenge_response_expected);

    if (memcmp(challenge_response_got, challenge_response_expected, 24) == 0) {
      debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) AP ChallengeResponse got is valid.\n");
      *outsize = 0;
      
      // Authentication was successful.
      mydata->eapsuccess = TRUE;
    }
    else {
      debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) AP ChallengeResponse got is NOT valid.\n");
      *outsize = -1;
      return XELEAP;
    }

    if (challenge_response_got != NULL) free(challenge_response_got);

    // We were successful, so generate keying material.

    ntPwdHash(MD4Hash, userdata->password);
    md4_calc(MD4HashHash, MD4Hash, 16);
    debug_printf(DEBUG_AUTHTYPES, "leap_session_key : ");
    debug_hex_printf(DEBUG_AUTHTYPES, MD4HashHash, 16);

    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) leapchallenges->pc : ");
    debug_hex_printf(DEBUG_AUTHTYPES, (uint8_t *)leapchallenges->pc, 8);

    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) leapchallenges->pr : ");
    debug_hex_printf(DEBUG_AUTHTYPES, (uint8_t *)leapchallenges->pr, 24);

    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) leapchallenges->apc : ");
    debug_hex_printf(DEBUG_AUTHTYPES, (uint8_t *)leapchallenges->apc, 8);

    debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) leapchallenges->apr : ");
    debug_hex_printf(DEBUG_AUTHTYPES, (uint8_t *)leapchallenges->apr, 24);

    GetMasterLEAPKey((char *)MD4HashHash, (char *) leapchallenges->apc, (char *) leapchallenges->apr, (char *) leapchallenges->pc, (char *) leapchallenges->pr, (char *)&MasterKey);
    debug_printf(DEBUG_AUTHTYPES, "MasterLEAPKey : ");
    debug_hex_printf(DEBUG_AUTHTYPES, (unsigned char *)&MasterKey, 16);

    // Finally, populate our thisint->keyingMaterial.
    if (mydata->keyingMaterial != NULL) {
      free(mydata->keyingMaterial);
      mydata->keyingMaterial = NULL;
    }
    mydata->keyingMaterial = (char *)malloc(64);  // 32 bytes each.
    if (mydata->keyingMaterial == NULL)
      return XEMALLOC;

    bzero(mydata->keyingMaterial, 64);
    memcpy(&mydata->keyingMaterial[32], &MasterKey, 16);
    memcpy(mydata->keyingMaterial, &MasterKey, 16);

  break;
  }    

  return XENONE;
}

/*******************************************************
 *
 * Assign our keying material.  (Return -1 if we can't generate keys.)
 *
 *******************************************************/
int eapleap_get_keys(struct interface_data *thisint)
{
  struct leap_data *mydata;
  struct config_network *network_data;

  // If we return keys, we return 0.  If we don't, return -1;

  if (!xsup_assert((thisint != NULL), "thisint != NULL", FALSE))
    return XEMALLOC;

  network_data = config_get_network_config();

  if (!xsup_assert((network_data != NULL), "network_data != NULL", FALSE))
    return XEMALLOC;

  if (!xsup_assert((network_data->activemethod != NULL),
		   "network_data->activemethod != NULL", FALSE))
    return XEMALLOC;

  mydata = (struct leap_data *)network_data->activemethod->eap_data;

  // Right now, we don't return anything from LEAP.
  thisint->keyingMaterial = (uint8_t *) mydata->keyingMaterial;
  thisint->keyingLength = 16;  // We only use 16 bytes for this keying material!

  return 0;
}

/*******************************************************
 *
 * Return if we have successfully authenticated.
 *
 *******************************************************/
int eapleap_done(struct generic_eap_data *thisint)
{
  struct leap_data *mydata;

  if (!xsup_assert((thisint != NULL), "thisint != NULL", FALSE))
    return XEMALLOC;

  mydata = (struct leap_data *)thisint->eap_data;

  if (!xsup_assert((mydata != NULL), "mydata != NULL", FALSE))
    return XEMALLOC;

  return mydata->eapsuccess;
}

/*******************************************************
 *
 * Clean up after ourselves.  This will get called when we get a packet that
 * needs to be processed requests a different EAP type.  It will also be
 * called on termination of the program.
 *
 *******************************************************/
int eapleap_cleanup(struct generic_eap_data *thisint)
{
  struct leap_data *mydata;

  if (!xsup_assert((thisint != NULL), "thisint != NULL", FALSE))
    return XEMALLOC;

  if (!xsup_assert((thisint->eap_data != NULL), "thisint->eap_data != NULL",
		   FALSE))
    return XEMALLOC;

  mydata = (struct leap_data *)thisint->eap_data;

  if (mydata->keyingMaterial != NULL)
    {
      free(mydata->keyingMaterial);
    }

  free(mydata);

  debug_printf(DEBUG_AUTHTYPES, "(EAP-LEAP) Cleaning up.\n");
  return XENONE;
}

/********************************************************
 *
 * We failed authentication for some reason, so clear out our password so
 * that we are prompted again at a later time.
 *
 ********************************************************/
int eapleap_failed(struct generic_eap_data *thisint)
{
  struct config_eap_leap *userdata;

  if (!xsup_assert((thisint != NULL), "thisint != NULL", FALSE))
    return XEMALLOC;

  if (!xsup_assert((thisint->eap_conf_data != NULL),
		   "thisint->eap_conf_data != NULL", FALSE))
    return XEMALLOC;

  userdata = (struct config_eap_leap *)thisint->eap_conf_data;

#ifndef NO_PWD_RESET
  /*
  if (userdata->password != NULL)
    {
      free(userdata->password);
      userdata->password = NULL;
    }
  */
#endif

  return XENONE;
}
